/*
 * Copyright (c) 2023 Xilinx
 * This package is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.

 * This package is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.

 * You should have received a copy of the GNU Lesser General Public
 * License along with this package; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "gstamadec.h"
#include "gstamacommon.h"
#include "gstxmaframeallocator.h"
#include "gstxrm.h"

#include <xmaerror.h>

GST_DEBUG_CATEGORY_STATIC (gst_ama_dec_debug);
#define GST_CAT_DEFAULT gst_ama_dec_debug

#define GST_AMA_DEC_LOW_LATENCY_DEFAULT XMA_DEC_LOW_LATENCY_DEFAULT
#define GST_AMA_DEC_LATENCY_LOGGING_DEFAULT XMA_DEC_LATENCY_LOGGING_DEFAULT

enum
{
  PROP_0,
  PROP_DEVICE,
  PROP_LOW_LATENCY,
  PROP_LATENCY_LOGGING,
};

static GstStaticPadTemplate src_template = GST_STATIC_PAD_TEMPLATE ("src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS (
        "video/x-raw(" GST_CAPS_FEATURE_MEMORY_AMA_MEMORY "), " \
            "format = (string) { NV12_4L4, NV12, I420, NV12_10LE32, P010_10LE, I420_10LE, RGBP, NV12_10LE_4L4 }," \
            "width = (int) [ 1, MAX ], " \
            "height = (int) [ 1, MAX ], " \
            "framerate = (fraction) [ 0/1, MAX ]"
    ));

typedef struct
{
  gint32 device;
  GstXma *xma;
  GstXrm *xrm;
  XmaDecoderSession *decoder_session;

  GstXMAFrameAllocator *allocator;

  gsize frame_size;

  int32_t low_latency;
  int32_t latency_logging;

  gboolean src_eos;
} GstAMADecPrivate;

G_DEFINE_TYPE_WITH_PRIVATE (GstAMADec, gst_ama_dec, GST_TYPE_VIDEO_DECODER);

static gboolean
gst_ama_dec_start (GstVideoDecoder * decoder)
{
  GstAMADec *self = GST_AMA_DEC (decoder);
  GstAMADecPrivate *priv = gst_ama_dec_get_instance_private (self);

  priv->xma = gst_xma_get_for_device (priv->device);
  if (!priv->xma) {
    GST_ELEMENT_ERROR (self, LIBRARY, INIT,
        ("Failed to initialize XMA library"), (NULL));
    return FALSE;
  }

  priv->allocator = GST_XMA_FRAME_ALLOCATOR (gst_xma_frame_allocator_new ());
  priv->src_eos = FALSE;

  return TRUE;
}

static gboolean
gst_ama_dec_close_session (GstAMADec * self)
{
  GstAMADecPrivate *priv = gst_ama_dec_get_instance_private (self);

  if (priv->decoder_session) {
    if (xma_dec_session_destroy (priv->decoder_session) != XMA_SUCCESS) {
      GST_ELEMENT_ERROR (self, LIBRARY, SHUTDOWN,
          ("Failed to destroy decoder session"), (NULL));
      return FALSE;
    }
    priv->decoder_session = NULL;
  }

  g_clear_pointer (&priv->xrm, gst_xrm_free);

  return TRUE;
}

static gboolean
gst_ama_dec_stop (GstVideoDecoder * decoder)
{
  GstAMADec *self = GST_AMA_DEC (decoder);
  GstAMADecPrivate *priv = gst_ama_dec_get_instance_private (self);

  if (!gst_ama_dec_close_session (self)) {
    return FALSE;
  }

  g_clear_object (&priv->xma);
  g_clear_object (&priv->allocator);

  return TRUE;
}

static gboolean
gst_ama_dec_set_format (GstVideoDecoder * decoder, GstVideoCodecState * state)
{
  GstAMADec *self = GST_AMA_DEC (decoder);
  GstAMADecPrivate *priv = gst_ama_dec_get_instance_private (self);
  GstAMADecClass *klass = GST_AMA_DEC_GET_CLASS (self);

  GstCaps *src_allowed_caps;
  guint bit_depth = 0;
  const GValue *val;
  XmaDecoderProperties dec_props = {0};
  XmaParameter params[3];
  GstVideoFormat output_format = GST_VIDEO_FORMAT_UNKNOWN;
  int32_t xma_output_format;
  GstVideoCodecState *output_state;
  guint i;

  if (!gst_ama_dec_close_session (self)) {
    return FALSE;
  }

  for (i = 0 ; i < gst_caps_get_size(state->caps); i++) {
    GstStructure *s = gst_caps_get_structure (state->caps, i);
    guint bit_depth_luma;
    guint bit_depth_chroma;

    if (!gst_structure_get_uint (s, "bit-depth-luma", &bit_depth_luma) ||
        !gst_structure_get_uint (s, "bit-depth-chroma", &bit_depth_chroma)) {
      GST_ELEMENT_ERROR (self, STREAM, FORMAT, NULL,
          ("Unable to detect input video bit depth"));
      return FALSE;
    }

    if (bit_depth_luma != bit_depth_chroma) {
      GST_ELEMENT_ERROR (self, STREAM, FORMAT, NULL,
          ("Different luma and chroma bit depths not supported"));
      return FALSE;
    }

    if (!bit_depth) {
      bit_depth = bit_depth_luma;
    } else if (bit_depth != bit_depth_luma) {
      GST_ELEMENT_ERROR (self, STREAM, FORMAT, NULL,
          ("Different bit depths types detected in sink pad's allowed caps."));
      return FALSE;
    }
  }

  if (!bit_depth) {
    GST_ELEMENT_ERROR (self, STREAM, FORMAT, NULL,
        ("Unable to detect input video bit depth"));
    return FALSE;
  }

  if (bit_depth != 8 && bit_depth != 10) {
    GST_ELEMENT_ERROR (self, STREAM, FORMAT, NULL,
        ("Unsupported bit depth %u", bit_depth));
    return FALSE;
  }

  GST_DEBUG_OBJECT (self, "sink caps: %" GST_PTR_FORMAT, state->caps);

  klass->configure_decoder (self, &dec_props);

  src_allowed_caps = gst_pad_get_allowed_caps (GST_VIDEO_ENCODER_SRC_PAD (decoder));

  GST_DEBUG_OBJECT (self, "src allowed caps: %" GST_PTR_FORMAT, src_allowed_caps);

  val = gst_structure_get_value (gst_caps_get_structure (src_allowed_caps, 0),
      "format");
  if (G_VALUE_HOLDS_STRING (val)) {
    output_format = gst_video_format_from_string (g_value_get_string (val));
  } else if (GST_VALUE_HOLDS_LIST (val)) {
    /* Format not fixed yet, pick the first supported format. */
    guint i;
    for (i = 0; i != gst_value_list_get_size (val); ++i) {
      output_format = gst_video_format_from_string (
          g_value_get_string (gst_value_list_get_value (val, i)));

      if (gst_video_format_get_info (output_format)->bits == bit_depth) {
        break;
      }
    }
  } else {
    output_format = (bit_depth == 8) ?
        GST_VIDEO_FORMAT_NV12_4L4 : GST_VIDEO_FORMAT_NV12_10LE32;
  }

  gst_clear_caps (&src_allowed_caps);

  if (output_format == GST_VIDEO_FORMAT_UNKNOWN) {
    GST_ELEMENT_ERROR (self, STREAM, FORMAT, NULL,
        ("Unable to select output video format"));
    return FALSE;
  }

  xma_output_format = gst_xma_get_format (output_format,
      (gint32 *) &dec_props.flags);

  params[0].name = XMA_DEC_OUTPUT_FORMAT;
  params[0].type = XMA_INT32;
  params[0].length = sizeof (int32_t);
  params[0].value = &xma_output_format;

  params[1].name = XMA_DEC_PARAM_LOW_LATENCY;
  params[1].type = XMA_INT32;
  params[1].length = sizeof (int32_t);
  params[1].value = &priv->low_latency;

  params[2].name = XMA_DEC_PARAM_LATENCY_LOGGING;
  params[2].type = XMA_INT32;
  params[2].length = sizeof (int32_t);
  params[2].value = &priv->latency_logging;

  dec_props.handle         = priv->xma->xma;
  dec_props.bits_per_pixel = bit_depth;
  dec_props.width          = GST_VIDEO_INFO_WIDTH (&state->info);
  dec_props.height         = GST_VIDEO_INFO_HEIGHT (&state->info);
  dec_props.param_cnt      = 3;
  dec_props.params         = params;

  priv->xrm = gst_xrm_reserve_dec_resources (priv->xma->device, dec_props.width,
      dec_props.height, GST_VIDEO_INFO_FPS_N (&state->info),
      GST_VIDEO_INFO_FPS_D (&state->info));
  if (!priv->xrm) {
    GST_ELEMENT_ERROR (self, RESOURCE, FAILED,
        ("Unable to reserve XRM resources for decoding %dx%d@%d/%d",
        dec_props.width, dec_props.height, GST_VIDEO_INFO_FPS_N (&state->info),
        GST_VIDEO_INFO_FPS_D (&state->info)), (NULL));
    return FALSE;
  }

  priv->decoder_session = xma_dec_session_create (&dec_props);

  output_state = gst_video_decoder_set_output_state (GST_VIDEO_DECODER (self),
      output_format, GST_VIDEO_INFO_WIDTH (&state->info),
      GST_VIDEO_INFO_HEIGHT (&state->info), state);

  priv->frame_size = GST_VIDEO_INFO_SIZE (&output_state->info);

  g_clear_pointer (&output_state, gst_video_codec_state_unref);

  return TRUE;
}

static XmaDataBuffer *
gst_ama_dec_copy_data_to_xma_buffer (GstAMADec * self, GstBuffer * inbuf)
{
  GstAMADecPrivate *priv = gst_ama_dec_get_instance_private (self);

  GstMapInfo map_info = GST_MAP_INFO_INIT;
  XmaDataBuffer *buf;

  if (!gst_buffer_map (inbuf, &map_info, GST_MAP_READ)) {
    GST_ERROR_OBJECT (self, "Failed to map input buffer!");
    return NULL;
  }

  buf = xma_data_buffer_alloc (priv->xma->xma, map_info.size, FALSE);

  memcpy (buf->data.buffer, map_info.data, map_info.size);

  gst_buffer_unmap (inbuf, &map_info);

  return buf;
}

static gint
_find_frame (gconstpointer pframe, gconstpointer pts)
{
  GstVideoCodecFrame *frame = (GstVideoCodecFrame *) pframe;
  GstClockTime ts;

  ts = GST_CLOCK_TIME_IS_VALID (frame->pts) ? frame->pts : frame->dts;

  return ts != *(GstClockTime *) pts;
}

static GstVideoCodecFrame *
gst_ama_dec_get_frame (GstAMADec * self, GstClockTime pts)
{
  GstVideoCodecFrame *frame = NULL;

  if (pts != 0) {
    GList *frames;
    GList *l;

    frames = gst_video_decoder_get_frames (GST_VIDEO_DECODER (self));
    l = g_list_find_custom (frames, &pts, _find_frame);
    if (l) {
      frame = gst_video_codec_frame_ref (l->data);
    }

    g_list_free_full (frames, (GDestroyNotify) gst_video_codec_frame_unref);
  }

  if (!frame) {
    /* The very first frame, no pts specified, or corresponding frame not found;
     * just return the oldest one. */
    frame = gst_video_decoder_get_oldest_frame (GST_VIDEO_DECODER (self));
  }

  return frame;
}

static GstFlowReturn
gst_ama_dec_receive_xma_frame (GstAMADec * self)
{
  GstAMADecPrivate *priv = gst_ama_dec_get_instance_private (self);

  GstFlowReturn result = GST_FLOW_ERROR;
  GstVideoCodecFrame *frame;
  GstVideoCodecState *state = NULL;
  GstMemory *mem;
  GstXMAFrame *xmaframe;
  int32_t xmares;

  static XmaFrameProperties frame_props = {
      XMA_VPE_FMT_TYPE, XMA_NV12_FMT_TYPE, 0, 0
  };

  mem = gst_xma_frame_allocator_alloc (priv->allocator, priv->frame_size,
      priv->xma, &frame_props, TRUE);

  xmaframe = GST_XMA_FRAME_CAST (mem);

  do {
    xmares = xma_dec_session_recv_frame (priv->decoder_session, xmaframe->frame);
    if (xmares == XMA_RESEND_AND_RECV) {
      g_usleep (100);
    }
  } while (xmares == XMA_RESEND_AND_RECV);

  if (xmares != XMA_SUCCESS) {
    if (xmares == XMA_EOS) {
      GST_LOG_OBJECT (self, "Received EOS from XMA");
      result = GST_FLOW_EOS;
    } else {
      GST_ELEMENT_ERROR (self, CORE, FAILED, (NULL),
          ("Could not receive decoded frame: %d", xmares));
    }
    goto out;
  }

  GST_LOG_OBJECT (self, "Received XMA frame with PTS: %" GST_TIME_FORMAT,
      GST_TIME_ARGS (xmaframe->frame->pts));
  xmaframe->frame->is_idr = 0;

  frame = gst_ama_dec_get_frame (self, xmaframe->frame->pts);
  if (!frame) {
    GST_WARNING_OBJECT (self, "XMA frame has no corresponding codec frame");
    result = GST_FLOW_OK;
    goto out;
  }

  if (G_UNLIKELY (priv->src_eos)) {
    gst_video_decoder_release_frame (GST_VIDEO_DECODER (self), frame);
    result = GST_FLOW_EOS;
  } else {
    frame->output_buffer = gst_buffer_new ();
    gst_buffer_append_memory (frame->output_buffer, g_steal_pointer (&mem));

    result = gst_video_decoder_finish_frame (GST_VIDEO_DECODER (self), frame);
    if (result == GST_FLOW_EOS) {
      GST_LOG_OBJECT (self, "Received EOS from downstream");
      priv->src_eos = TRUE;
    }
  }

out:
  g_clear_pointer (&mem, gst_memory_unref);
  g_clear_pointer (&state, gst_video_codec_state_unref);

  return result;
}

static GstFlowReturn
gst_ama_dec_handle_frame (GstVideoDecoder * decoder,
    GstVideoCodecFrame * frame)
{
  GstAMADec *self = GST_AMA_DEC (decoder);
  GstAMADecPrivate *priv = gst_ama_dec_get_instance_private (self);

  XmaDataBuffer *buf;
  int32_t xmares;
  int32_t data_used;

  buf = gst_ama_dec_copy_data_to_xma_buffer (self, frame->input_buffer);
  if (!buf) {
    return GST_FLOW_ERROR;
  }

  buf->pts = GST_CLOCK_TIME_IS_VALID (frame->pts) ?
      frame->pts : XMA_AV_NOPTS_VALUE;

  GST_LOG_OBJECT (self, "Sending frame with PTS: %" GST_TIME_FORMAT,
      GST_TIME_ARGS (buf->pts));

  xmares = xma_dec_session_send_data (priv->decoder_session,
      g_steal_pointer (&buf), &data_used);

  if (xmares != XMA_SUCCESS && xmares != XMA_SEND_MORE_DATA) {
    GST_ELEMENT_ERROR (self, CORE, FAILED, (NULL), ("failed to decode frame"));
    goto error;
  }

  gst_video_codec_frame_unref (frame);

  if (xmares == XMA_SEND_MORE_DATA) {
    return GST_FLOW_OK;
  }

  return gst_ama_dec_receive_xma_frame (self);

error:
  gst_video_decoder_drop_frame (decoder, frame);
  return GST_FLOW_ERROR;
}

static GstFlowReturn
gst_ama_dec_finish (GstVideoDecoder * decoder)
{
  GstAMADec *self = GST_AMA_DEC (decoder);
  GstAMADecPrivate *priv = gst_ama_dec_get_instance_private (self);

  GstFlowReturn res;
  int32_t xmares;
  guint count = 0;

  do {
    xmares = xma_dec_session_send_data (priv->decoder_session, NULL, NULL);
    if (xmares == XMA_TRY_AGAIN) {
      if (++count >= 1000) {
        xmares = XMA_ERROR;
        break;
      }
      g_usleep (100);
    }
  } while (xmares == XMA_TRY_AGAIN);


  if (xmares != XMA_SUCCESS) {
    GST_ELEMENT_ERROR (self, CORE, FAILED, (NULL),
        ("could not send EOS to decoder session"));
  }

  do {
    res = gst_ama_dec_receive_xma_frame (self);
  } while (res == GST_FLOW_OK);

  return res;
}

static gboolean
gst_ama_dec_negotiate (GstVideoDecoder * decoder)
{
  GstAMADec *self = GST_AMA_DEC (decoder);
  GstAMADecPrivate *priv = gst_ama_dec_get_instance_private (self);

  GstVideoCodecState *state;

  state = gst_video_decoder_get_output_state (decoder);
  if (state && !state->caps) {
    state->caps = gst_video_info_to_caps (&state->info);
    gst_caps_set_features_simple (state->caps,
        gst_caps_features_from_string (GST_CAPS_FEATURE_MEMORY_AMA_MEMORY));
    if (priv->xma) {
      gst_ama_device_add_to_caps (state->caps, priv->xma->device);
    }

    gst_video_codec_state_unref (state);
  }

  return GST_VIDEO_DECODER_CLASS (gst_ama_dec_parent_class)->negotiate (decoder);
}

static void
gst_ama_dec_init (GstAMADec * dec)
{
}

static void
gst_ama_dec_set_property (GObject * object, guint prop_id, const GValue * value,
    GParamSpec * pspec)
{
  GstAMADec *self = GST_AMA_DEC (object);
  GstAMADecPrivate *priv = gst_ama_dec_get_instance_private (self);

  switch (prop_id) {
    case PROP_DEVICE:
      priv->device = g_value_get_int (value);
      break;
    case PROP_LOW_LATENCY:
      priv->low_latency = g_value_get_boolean (value) ? 1 : 0;
      break;
    case PROP_LATENCY_LOGGING:
      priv->latency_logging = g_value_get_boolean (value) ? 1 : 0;
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_ama_dec_get_property (GObject * object, guint prop_id, GValue * value,
    GParamSpec * pspec)
{
  GstAMADec *self = GST_AMA_DEC (object);
  GstAMADecPrivate *priv = gst_ama_dec_get_instance_private (self);

  switch (prop_id) {
    case PROP_DEVICE:
      g_value_set_int (value, priv->device);
      break;
    case PROP_LOW_LATENCY:
      g_value_set_boolean (value, priv->low_latency);
      break;
    case PROP_LATENCY_LOGGING:
      g_value_set_boolean (value, priv->latency_logging);
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
      break;
  }
}

static void
gst_ama_dec_class_init (GstAMADecClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
  GstElementClass *element_class = GST_ELEMENT_CLASS (klass);
  GstVideoDecoderClass *decoder_class = GST_VIDEO_DECODER_CLASS (klass);

  gobject_class->set_property = gst_ama_dec_set_property;
  gobject_class->get_property = gst_ama_dec_get_property;

  decoder_class->start = gst_ama_dec_start;
  decoder_class->stop = gst_ama_dec_stop;
  decoder_class->set_format = gst_ama_dec_set_format;
  decoder_class->handle_frame = gst_ama_dec_handle_frame;
  decoder_class->finish = gst_ama_dec_finish;
  decoder_class->negotiate = gst_ama_dec_negotiate;

  g_object_class_install_property (gobject_class, PROP_DEVICE,
      g_param_spec_int ("device", "XMA device",
          "Numeric ID of an XMA device (-1 = auto)",
          -1, G_MAXINT32, GST_XMA_DEVICE_AUTO,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_LOW_LATENCY,
      g_param_spec_boolean ("low-latency", "Low latency",
          "Whether to enable low latency decoding",
          GST_AMA_DEC_LOW_LATENCY_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  g_object_class_install_property (gobject_class, PROP_LATENCY_LOGGING,
      g_param_spec_boolean ("latency-logging", "Latency logging",
          "Log latency info to syslog",
          GST_AMA_DEC_LATENCY_LOGGING_DEFAULT,
          G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS | G_PARAM_CONSTRUCT));

  gst_element_class_add_static_pad_template (element_class, &src_template);

  GST_DEBUG_CATEGORY_INIT (gst_ama_dec_debug, "ama_dec", 0,
      "Xilinx AMA H264/H265/AV1 decoder");
}
